/* 
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 1049 $ $LastChangedDate: 2002-10-27 15:32:45 +0100 (Sun, 27 Oct 2002) $
 */

#include <gont/runtime.h>
#include <gont/array.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <errno.h>

gont_int Ipc__dup(gont_int fd)
{
	int ret;

	ret = dup(fd);

	if (ret == -1)
		gont_sys_error(errno);
	
	return ret;
}

void Ipc__dup2(gont_int fd1, gont_int fd2)
{
	if (dup2(fd1, fd2) == -1)
		gont_sys_error(errno);
}

gont_int Ipc__fork(void)
{
	return fork();
}

gont_int Ipc__getpid(void)
{
	return getpid();
}

gont_int Ipc__getppid(void)
{
	return getppid();
}

void **Ipc__pipe(void)
{
	int p[2];

	if (pipe(p))
		gont_sys_error(errno);
	
	return gont_make_tuple(2, (gont_int)p[0], (gont_int)p[1]);
}

gont_int Ipc___execve(const char *filename, struct gont_array *argv, 
		      struct gont_array *envp)
{
	return execve(filename, (char **)argv->data, (char **)envp->data);
}

void **Ipc___waitpid(gont_int pid, gont_int options)
{
	int opts = 0, status = 0;
	gont_int ret;
	
	if (options & 1)
		opts |= WNOHANG;
	if (options & 2)
		opts |= WUNTRACED;
	
	ret = waitpid(pid, &status, opts);

	if (ret == 0)
		return gont_make_tuple(3, ret, (gont_int)0, (gont_int)0);
	
	if (ret == (gont_int)-1)
		gont_sys_error(errno);
	
	if (WIFEXITED(status))
		return gont_make_tuple(3, ret, (gont_int)1, 
				       (gont_int)(WEXITSTATUS(status)));

#ifdef WCOREDUMP
	if (WIFSIGNALED(status) && WCOREDUMP(status))
		return gont_make_tuple(3, ret, (gont_int)3, 
				       (gont_int)(WTERMSIG(status)));
#endif
	
	if (WIFSIGNALED(status))
		return gont_make_tuple(2, ret, (gont_int)2, 
				       (gont_int)(WTERMSIG(status)));
				       
	if (WIFSTOPPED(status))
		return gont_make_tuple(3, ret, (gont_int)4,
				       (gont_int)(WSTOPSIG(status)));

	/* What the hack? */
	return gont_make_tuple(3, ret, (gont_int)5, (gont_int)0);
}

