/* 
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 931 $ $LastChangedDate: 2002-09-11 10:32:30 +0200 (Wed, 11 Sep 2002) $
 *
 * Standard library for Gont, Unitype module C implementation.
 */

#include <gont/runtime.h>

enum {
	Letter_flag		= 0x0100,
	Number_flag		= 0x0200,
	Punctation_flag		= 0x0400,
	Symbol_flag		= 0x0800,
	Separator_flag		= 0x1000,
	Mark_flag		= 0x2000
};

enum {
	Letter_uppercase = 1,
	Letter_lowercase,
	Letter_titlecase,
	Letter_modifier,
	Letter_other,
	Mark_non_spacing,
	Mark_space_combining,
	Mark_enclosing,
	Number_decimal_digit,
	Number_letter,
	Number_other,
	Punctation_connector,
	Punctation_dash,
	Punctation_open,
	Punctation_close,
	Punctation_initial,
	Punctation_final,
	Punctation_other,
	Symbol_math,
	Symbol_currency,
	Symbol_modifier,
	Symbol_other,
	Separator_space,
	Separator_line,
	Separator_paragraph,
	Other_control,
	Other_format,
	Other_surrogate,
	Other_private_use,
	Other_not_assigned,
	Unknown
};

static struct gont_union unions[] = {
	{0, 0},
	{Letter_uppercase, 0},
	{Letter_lowercase, 0},
	{Letter_titlecase, 0},
	{Letter_modifier, 0},
	{Letter_other, 0},
	{Mark_non_spacing, 0},
	{Mark_space_combining, 0},
	{Mark_enclosing, 0},
	{Number_decimal_digit, 0},
	{Number_letter, 0},
	{Number_other, 0},
	{Punctation_connector, 0},
	{Punctation_dash, 0},
	{Punctation_open, 0},
	{Punctation_close, 0},
	{Punctation_initial, 0},
	{Punctation_final, 0},
	{Punctation_other, 0},
	{Symbol_math, 0},
	{Symbol_currency, 0},
	{Symbol_modifier, 0},
	{Symbol_other, 0},
	{Separator_space, 0},
	{Separator_line, 0},
	{Separator_paragraph, 0},
	{Other_control, 0},
	{Other_format, 0},
	{Other_surrogate, 0},
	{Other_private_use, 0},
	{Other_not_assigned, 0},
	{Unknown, 0}
};

#define Letter_uppercase_F	(Letter_uppercase | Letter_flag)
#define Letter_lowercase_F	(Letter_lowercase | Letter_flag)
#define Letter_titlecase_F	(Letter_titlecase | Letter_flag)
#define Letter_modifier_F	(Letter_modifier | Letter_flag)
#define Letter_other_F		(Letter_other | Letter_flag)
#define Mark_non_spacing_F	(Mark_non_spacing | Mark_flag)
#define Mark_space_combining_F	(Mark_space_combining | Mark_flag)
#define Mark_enclosing_F	(Mark_enclosing | Mark_flag)
#define Number_decimal_digit_F	(Number_decimal_digit | Number_flag)
#define Number_letter_F		(Number_letter | Number_flag)
#define Number_other_F		(Number_other | Number_flag)
#define Punctation_connector_F	(Punctation_connector | Punctation_flag)
#define Punctation_dash_F	(Punctation_dash | Punctation_flag)
#define Punctation_open_F	(Punctation_open | Punctation_flag)
#define Punctation_close_F	(Punctation_close | Punctation_flag)
#define Punctation_initial_F	(Punctation_initial | Punctation_flag)
#define Punctation_final_F	(Punctation_final | Punctation_flag)
#define Punctation_other_F	(Punctation_other | Punctation_flag)
#define Symbol_math_F		(Symbol_math | Symbol_flag)
#define Symbol_currency_F	(Symbol_currency | Symbol_flag)
#define Symbol_modifier_F	(Symbol_modifier | Symbol_flag)
#define Symbol_other_F		(Symbol_other | Symbol_flag)
#define Separator_space_F	(Separator_space | Separator_flag)
#define Separator_line_F	(Separator_line | Separator_flag)
#define Separator_paragraph_F	(Separator_paragraph | Separator_flag)
#define Other_control_F		(Other_control)
#define Other_format_F		(Other_format)
#define Other_surrogate_F	(Other_surrogate)
#define Other_private_use_F	(Other_private_use)

#define U(cat, props) ((props << 16) | cat ## _F)

#include "c_unitype.inc"

#define lookup(ch) \
	(((ch)>>8) < sizeof(pages) / sizeof(pages[0]) && pages[(ch)>>8] ? \
	 pages[(ch)>>8][(ch) & 0xff] : 0)

#define CAT_MASK 0xff
#define cat(ch) (lookup(ch) & CAT_MASK)

#define OFFSET(c)  (c >> 16)

struct gont_union *Unitype__get_category(gont_int ch)
{
	int c = lookup(ch);

	if (c)
		return &unions[c & CAT_MASK];
	else
		return &unions[Other_not_assigned];
}

gont_int Unitype__is_uppercase(gont_int ch)
{
	return cat(ch) == Letter_uppercase;
}

gont_int Unitype__is_lowercase(gont_int ch)
{
	return cat(ch) == Letter_lowercase;
}

gont_int Unitype__is_letter(gont_int ch)
{
	return (lookup(ch) & Letter_flag) != 0;
}

gont_int Unitype__is_number(gont_int ch)
{
	return (lookup(ch) & Number_flag) != 0;
}

gont_int Unitype__is_alnum(gont_int ch)
{
	return (lookup(ch) & (Number_flag | Letter_flag)) != 0;
}

gont_int Unitype__is_punct(gont_int ch)
{
	return (lookup(ch) & Punctation_flag) != 0;
}

gont_int Unitype__is_control(gont_int ch)
{
	return cat(ch) == Other_control; 
}

gont_int Unitype__is_symbol(gont_int ch)
{
	return (lookup(ch) & Symbol_flag) != 0;
}

gont_int Unitype__is_mark(gont_int ch)
{
	return (lookup(ch) & Mark_flag) != 0;
}

gont_int Unitype__is_separator(gont_int ch)
{
	return (lookup(ch) & Separator_flag) != 0;
}

/* Conversions.  */

gont_int Unitype__to_uppercase(gont_int ch)
{
	int c = lookup(ch);

	if ((c & CAT_MASK) == Letter_lowercase)
		return ch + OFFSET(c);
	else
		return ch;
}

gont_int Unitype__to_lowercase(gont_int ch)
{
	int c = lookup(ch);

	if ((c & CAT_MASK) == Letter_uppercase)
		return ch + OFFSET(c);
	else
		return ch;
}

gont_int Unitype__get_digit_value(gont_int ch)
{
	int c = lookup(ch);

	if ((c & CAT_MASK) == Number_decimal_digit)
		return OFFSET(c);
	else {
		gont_invalid_argument("Unitype::get_digit_value");
		return 0;
	}
}
