/* 
 * Copyright (c) 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 1012 $ $LastChangedDate: 2002-09-21 20:16:14 +0200 (Sat, 21 Sep 2002) $
 *
 * Runtime environment for Gont, Core module interface.
 * 
 * This module is opened at the beginning of every compilation unit.
 */
 
/** Return decimal string representation of [i].  */
string itoa(int t) = "Core__itoa";

/** Convert [s] from decimal string to integer.  If [s] is not valid decimal
    integer with optional white spaces at the beginning and at the end
    [Invalid_argument] is raised.  This is different from atoi(3), 
    for example in C [atoi("3ble") == 3], in Gont exception is raised.  */
int atoi(string s) = "Core__atoi";

/** Return decimal string representation of [f].  */
string ftoa(float f) = "Core__ftoa";

/** Convert [s] from decimal string to float.  If [s] is not valid floating
    point number (as with strtod(3)), with optional white spaces at the
    beginning and at the end [Invalid_argument] is raised.  */
float atof(string) = "Core__atof";

/** Return either "true" iff [b == true] or "false" otherwise.  */
string string_of_bool(bool b);

/** Return [true] iff [s == "true"], [false] iff [s == "false"], raise
    [Invalid_argument] otherwise.  */
bool bool_of_string(string s);

/** An alias for Core::itoa.  */
string int_of_string(int) = "Core__itoa";

/** An alias for Core::atoi.  */
int string_of_int(string) = "Core__atoi";

/** An alias for Core::ftoa.  */
string string_of_float(float) = "Core__ftoa";

/** An alias for Core::atof.  */
float float_of_string(string) = "Core__atof";

/** [f] is ignored.  It is useful if you want discard return value of 
    function, and don't want warning, then you do: [ignore(f())].  
    Of course this is only relevant if you compile with -Wignored-retval.  */
void ignore('a f);

/** Return [f] partially evaluated on [c], i.e. [fun (x) f(c, x)].  */
*('a ('b)) partial(*('a ('c, 'b)) f, 'c c);

/** Return [f] partially evaluated on [c, d], i.e. [fun (x) f(c, d, x)].  */
*('a ('b)) partial2(*('a ('c, 'd, 'b)) f, 'c c, 'd d);

/** Print [s] on [stdout].  Note that it doesn't append ["\n"].  */
void puts(string s) = "Core__puts";

/** Print [s] on [stderr].  */
void puts_err(string s) = "Core__puts_err";

/** Read and return one line of input from [stdin].  */
string read_line() = "Core__read_line";

/** Read one word (sequence of non-white-space characters) from [stdin],
    and return it converted to int.  It's more or less the same as 
    [scanf("%i", &i);] in C.  */
int read_int();

/** Read one word (sequence of non-white-space characters) from [stdin],
    and return it converted to int.  It's more or less the same as 
    [scanf("%f", &f);] in C.  */
float read_float();

/** Read one word (sequence of non-white-space characters) from [stdin],
    and return it.  It's more or less the same as [scanf("%s", buf);] in C.  */
string read_string() = "Core__read_string";

/** Return first element of pair [p].  */
'a fst(*('a, 'b) p);

/** Return second element of pair [p].  */
'b snd(*('a, 'b) p);

/** Return unicode character number [x].  
    Raise [Invalid_argument] if [x] is outside allowed range (0 - 2^31-1). */
char char_of_int(int x);

/** Return unicode number of [x].  */
int int_of_char(char x) = "gont_id";

/** Type of immutable, ML-style lists.  See List module for operations on 
    lists.  */
union <'a>list {
    void Nil;
    *('a, <'a>list) Cons;
}

/** Type of mutable references.  */
struct <'a>ref {
    'a ref_contents;
}

/** Get value from reference [r].  */
'a ref_get(<'a>ref r);

/** Make reference out of [a].  */
<'a>ref ref_make('a data);

/** Set value in reference [r] to [a].  */
void ref_set(<'a>ref r, 'a a);

/** Option type.  */
union <'a>option {
	void None;
	'a Some;
}

/** Standard exceptions.  */

/** Raised when none of pattern clauses matched expression.  */
exception void Match_failure;

/** Raised upon attempt to access [opt_struct] that is [null].  */
exception void Null_access;

/** Raised by various library functions, when they are given arguments they
    can't handle.  */
exception string Invalid_argument;

/** Raised by library functions, that are undefined for given arguments.  */
exception string Failure;

/** Raised by library functions, when object to be returned was not to be 
    found.  */
exception void Not_found;

/** Raised by library functions that had problems with underlaying C functions, 
    This generally involves [strerror(errno)].  */
exception string Sys_error;

/** Raised by input functions, when end of file is reached.  */
exception void End_of_file;

/** Unicode exceptions.  */

/** Can be raised during conversion from some external encoding to string.  */
exception void Invalid_multibyte_sequence;

/** Can be raised during conversion from some external encoding to string.  */
exception void Incomplete_multibyte_sequence;

/** Can be raised uppon conversion from <char>Array::t to external encoding
    or string, when some character exceeds 2^31.  */
exception void Invalid_unicode_character;

/** Can be raised when converting string to some limited external encoding.
    Cannot be raised if converting to utf-*.  */
exception void Character_not_supported_by_encoding;

// vim:sts=4:sw=4
