/* 
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 893 $ $LastChangedDate: 2002-09-07 20:40:04 +0200 (Sat, 07 Sep 2002) $
 *
 * Standard library for Gont, Array module implementation.
 */

open Array;

extern <'a>t alloc(int sz) = "Array___alloc";
extern <'a>t _resize(<'a>t ar, int sz) = "Array___resize";

<'a>t init(int n, *('a (int)) f)
{
    let r = alloc(n);
    
    for (int i = 0; i < n; i++)
	set(r, i, f(i));
	
    return r;
}

<'a>t empty(int size_hint)
{
    let r = alloc(size_hint);
    _resize(r, 0);
    return r;
}

void clear(<'a>t a)
{
    _resize(a, 0);
}

<'a>t concat_sep(<'a>t sep, <<'a>t>list lst)
{
    int sep_len = length(sep);
    
    int count_len(int len, <<'a>t>list l)
    {
	switch l {
	case Cons(x, xs):
	    return count_len(len + sep_len + length(x), xs);
	case Nil:
	    return len - sep_len;
	}
    }
    
    int len = count_len(0, lst);
    
    if (len <= 0)
	return alloc(0);
    
    <'a>t r = alloc(len);

    void blt(int pos, <<'a>t>list l)
    {
	switch l {
	case Cons(x, Nil):
	    blit(x, 0, r, pos, length(x));
		
	case Cons(x, xs):
	    blit(x, 0, r, pos, length(x));
	    pos = pos + length(x);
	    blit(sep, 0, r, pos, sep_len);
	    pos = pos + sep_len;
	    blt(pos, xs);
		
	case Nil:
	    raise Failure("Array::concat: impossible");
	}
    }

    blt(0, lst);

    return r;
}

<'a>t concat(<<'a>t>list l)
{
    // FIXME: could be faster...
    return concat_sep(alloc(0), l);
}

<'a>t copy(<'a>t src)
{
    let l = length(src);
    let r = alloc(l);
    blit(src, 0, r, 0, l);
    return r;
}

void append_array(<'a>t dst, <'a>t src)
{
    let ld = length(dst);
    let ls = length(src);
    
    _resize(dst, ld + ls);
    blit(src, 0, dst, ld, ls);
}

void iter(*(void ('a)) f, <'a>t array)
{
    int sz = length(array);
    
    for (int i = 0; i < sz; i++)
	f(get(array, i));
}

<'a>list to_list(<'a>t array)
{
    <'a>list make(<'a>list ret, int pos) 
    {
	if (pos < 0)
	    return ret;
	else
	    return make(Cons(get(array, pos), ret), pos - 1);
    }

    return make(Nil, length(array) - 1);
}

<'a>t of_list(<'a>list lst)
{
    let r = alloc(List::length(lst));
    
    void init(<'a>list l, int pos) 
    {
	switch l {
	case Cons(x, xs):
	    set(r, pos, x);
	    init(xs, pos + 1);
	case Nil:
	    skip;
	}
    }
    
    init(lst, 0);
    
    return r;
}

<'a>t sub(<'a>t a, int off, int len)
{
    // check it here
    let l = length(a) in
    if (off < 0 || off > l || len < 0 || off + len > l)
	raise Invalid_argument("Array::sub");

    let r = alloc(len);
    blit(a, off, r, 0, len);

    return r;
}

// vim:sts=4:sw=4
