/* 
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 848 $ $LastChangedDate: 2002-09-01 21:12:20 +0200 (Sun, 01 Sep 2002) $
 *
 * Fifo - first-in/first-out queue.
 */

open Slist;

struct <'a>t {
    <'a>Slist::t head;
    <'a>Slist::t tail;
}

'a get(<'a>t q)
{
    if (q.head == null)
	raise Invalid_argument("Fifo::get");
	
    _ r = q.head.data;
    
    if (Hacks::same_ptrs(q.head, q.tail)) {
	q.head = null;
	q.tail = null;
    } else {
	q.head = q.head.next;
    }

    return r;
}

'a peek(<'a>t q)
{
    if (q.head == null)
	raise Invalid_argument("Fifo::peek");

    return q.head.data;
}

void put(<'a>t q, 'a a)
{
    q.head = {Slist::data = a, Slist::next = q.head};
    
    if (q.tail == null)
	q.tail = q.head;
}

<'a>t create()
{
    return {head = null, tail = null};
}

bool empty(<'a>t q)
{
    return q.head == null;
}

// vim:sts=4:sw=4
