// $Rev$ $LastChangedDate$

<string, string>Gontlex::mode m = Gontlex::make_mode();

open Fileutil;
open Stdio;
open Gontlex;

int err_cnt = 0;
Iconv::charset utf8 = Iconv::lookup_charset("utf-8");
<Xml::node>Stack::t modules = Stack::create();

struct entry {
    string desc;
    string name;
    string proto;
}

def xmlize_entry(e)
{
    def add_refs(buf)
    {
	_ last_end = 0;
	_ len = Array::length(buf);
	_ out = Stack::create();
	
	def is_id(c)
	    (Unitype::is_ascii_alnum(c) || c == '_');
	
	for (int i = 0; i < len - 1; i++) {
	    if (Array::get(buf, i) == ':' && Array::get(buf, i + 1) == ':') {
		_ beg = i;
		while (--beg >= 0 && is_id(Array::get(buf, beg)))
		    skip;
		beg++;
		string module, name;
		let a = Array::sub(buf, beg, i - beg) in
		module = String::of_char_array(a);
		
		_ end = i + 1;
		while (++end < len && is_id(Array::get(buf, end)))
		    skip;
		end--;
		let a = Array::sub(buf, i + 2, end - i - 1) in
		name = String::of_char_array(a);

		if (name != "" && module != "") {
		    let a = Array::sub(buf, last_end, beg - last_end) in
		    let s = String::of_char_array(a) in
		    Stack::push(out, Xml::Text(s));
		    
		    let a = [("name", name), ("module", module)] in
		    let c = [Xml::Text(module + "::" + name)] in
		    let ref = Xml::Element("ref", c, a) in
		    Stack::push(out, ref);

		    last_end = end + 1;
		    i = end;
		}
	    }
	}
	
	let a = Array::sub(buf, last_end, len - last_end) in
	let s = String::of_char_array(a) in
	Stack::push(out, Xml::Text(s));

	Array::clear(buf);

	return List::rev(Stack::to_list(out));
    }
    
    def add_tt(s)
    {
	_ buf = Array::empty(1000);
	_ lev = 0;
	_ out = Stack::create();

	def add_ch(c)
	{
	    if (c == '[') {
		lev++;
		if (lev == 1) {
		    List::iter(fun (e) (Stack::push(out, e)), add_refs(buf));
		    return;
		}
	    } else if (c == ']') {
		lev--;
		if (lev == 0) {
		    Stack::push(out, Xml::Element("tt", add_refs(buf), []));
		    return;
		}
	    }

	    Array::append(buf, c);
	}

	String::iter(add_ch, s);

	if (Array::length(buf) != 0)
	    List::iter(fun (e) (Stack::push(out, e)), add_refs(buf));

	return Xml::Element("desc", List::rev(Stack::to_list(out)), []);
    }

    let d = add_tt(e.desc);

    _ attrs;
    
    if (e.name == "")
	attrs = [];
    else
	attrs = [("name", e.name)];
    
    if (e.proto != "")
	let p = Xml::Element("proto", [Xml::Text(e.proto)], []) in
	return Xml::Element("refentry", [p, d], attrs);
    else
	return Xml::Element("general", [d], []);
}

def parse(lex)
{
    _ entries = Stack::create();
    
    def get()
    {
	try {
	    switch Gontlex::get_token(lex) {
	    case Eof:
		raise End_of_file;
	    case x:
		return x;
	    }
	} with {
	case Error(_):
	    return get();
	}
    }

    def non_blank()
    {
	switch get() {
	case Blank(_):
	    return non_blank();
	case x:
	    return x;
	}
    }

    def unlex(t)
    {
	switch t {
	case Id(s):
	case Blank(s):
	case Comment(s):
	case Operator(s):
	    return s;
	
	case Ty_var(a):
	    return "'" + a;

	// these are prolly not going to be used
	case String(s):
	    return "\"" + s + "\"";
	    
	case Char(c):
	    return "'" + String::make(1, c) + "'";
	
	case Int(s):
	case Float(s):
	case Keyword(s):
	    return s;
	
	case Eof:
	    return Sys::abort();
	}
    }

    def chomp_spaces(lst)
    {
	switch lst {
	case [Blank(_), Operator(";") as e]:
	    return [e];
	case Cons(hd, tl):
	    return Cons(hd, chomp_spaces(tl));
	case Nil:
	    return Nil;
	}
    }

    def aint_junk(t)
    {
	switch t {
	case Blank(_):
	case Comment(_):
	    return false;
	case _:
	    return true;
	}
    }

    def get_proto(lev)
    {
	_ t = get();
	
	switch t {
	case Operator("{"):
	    lev++;
	    
	case Operator("}"):
	    if (lev == 1)
		return [t];
	    else
		lev--;
	    
	case Operator("=") if lev == 0:
	    // skip '= "foobar"' part
	    _ p = get_proto(0);
	    def is_xn(lst)
	    {
		switch lst {
		case Cons(String(_), _):
		    return true;
		case _:
		    return false;
		}
	    }

	    if (is_xn(List::filter(aint_junk, p)))
		return [Operator(";")];
	    else
		return Cons(t, p);
	    
	case Operator(";") if lev == 0:
	    return [t];

	case _:
	    skip;
	}
	
	return Cons(t, get_proto(lev));
    }

    def find_name(lev, lst)
    {
	switch lst {
	case Cons(Id(i), Cons(Operator(";"), _)) if lev == 0:
	case Cons(Id(i), Cons(Operator("{"), _)) if lev == 0:
	case Cons(Id(i), Cons(Operator("("), _)) if lev == 0:
	    return i;
	
	case Cons(Operator("<"), tl):
	case Cons(Operator("{"), tl):
	case Cons(Operator("("), tl):
	    return find_name(lev + 1, tl);
	    
	case Cons(Operator(">"), tl):
	case Cons(Operator("}"), tl):
	case Cons(Operator(")"), tl):
	    return find_name(lev - 1, tl);
	
	case Cons(_, tl):
	    return find_name(lev, tl);
	
	case Nil:
	    raise Not_found;
	}
    }
    
    def process_comment(c)
    {
	if (String::sub(c, 0, 3) != "/**")
	    return;
	    
	// chop comment delimeters
	c = "   " + String::sub(c, 3, String::length(c) - 5);
	_ e = {desc = c, name = "", proto = ""};
	
	Stack::push(entries, e);

	_ t = non_blank();
	
	switch t {
	case Comment(c):
	    process_comment(c);
	
	case Id(_):
	case Ty_var(_):
	case Operator(_):
	    _ lst = Cons(t, chomp_spaces(get_proto(0)));
	    _ str = String::concat("", List::map(unlex, lst));
	    try {
		e.name = find_name(0, List::filter(aint_junk, lst));
	    } with {
	    case Not_found:
		report_error("cannot deduce name of `" + str + "'");
	    }
	    e.proto = str;
	    
	case _:
	    skip;
	}
    }

    try {
	for (;;) {
	    switch get() {
	    case Comment(c):
		process_comment(c);
	    case _:
		skip;
	    }
	}
    } with {
    case End_of_file:
	return List::rev(Stack::to_list(entries));
    }
}

def process_file(name)
{
    try {
	switch split_filename(basename(name)) {
	case (b, "gi"):
	case (b, "g"):
	    _ f = fopen(name, "r");
	    set_charset(f, utf8);
	    _ l = Gontlex::of_genlex(m, Genlex::of_stream(f));
	    _ entries = parse(l);
	    switch entries {
	    case Nil:
		skip;
	    case _:
		_ xml = Xml::Element("module", List::map(xmlize_entry, entries),
				     [("name", String::capitalize(b))]);
		Stack::push(modules, xml);
	    }

	case _:
	    report_error("don't know how to handle `" + name + "'");
	    err_cnt++;
	}
    } with {
    case Sys_error(e):
	report_error(name + ": " + e);
    }
}

void init_mode()
{
    _ ops = [",", "(", ")", ";", "=", "<", ">", "*", "{", "}", "::"];
    
    List::iter(fun (o) (Gontlex::add_operator(m, o, o)), ops);

    Gontlex::handle_tyvars(m);
    Gontlex::keep_blanks(m);
    Gontlex::keep_comments(m);
}

section init 
{
    init_mode();
    for (int i = 1; i < Array::length(Sys::argv); i++)
	process_file(Array::get(Sys::argv, i));

    if (err_cnt != 0)
	Sys::exit(1);
    
    _ xml = Xml::Element("modules", List::rev(Stack::to_list(modules)), []);
    
    Xml::save_to_file_indented(Xml::make_doc(xml), "modules.xml");
}

// vim:sts=4:sw=4
